package PreMigration;

use strict;
use warnings;
use Logging;

my @MESSAGES;

# severity levels
use constant {
  ERROR => 4,
  WARNING => 3,
  # migration is possible, but is not implemented in the migration agent, user should migrate objects manually
  NOT_IMPLEMENTED => 2,
  # use this level to show non-critical information
  INFO => 1,
  # use this level to show migration mapping
  MAPPING => 0
};

# names of severity levels
use constant SEVERITY_NAMES => {
  4 => 'ERROR',
  3 => 'WARNING',
  2 => 'NOT IMPLEMENTED',
  1 => 'INFO',
  0 => 'MAPPING',
};

# components
use constant {
  DNS => 'DNS',
  MAIL => 'Mail',
  WEB => 'Web', 
  DB => 'Database',
  CP => 'Control Panel',
};

use PreMigrationMessages;

# add message by id from PreMigrationMessages.pm
sub message {
  my ($id, $context, $owners) = @_;

  if (!$PreMigrationMessages::MESSAGES{$id}) {
    print STDERR "Pre-migration message '$id' does not exist";
    return;
  }

  my $messageInfo = $PreMigrationMessages::MESSAGES{$id};

  # substitute context variables
  my $briefMessage = $messageInfo->{'briefMessage'};
  my $detailedMessage = $messageInfo->{'detailedMessage'};
  foreach my $key (keys %{$context}) {
    $briefMessage =~ s/{$key}/$context->{$key}/g;
    $detailedMessage =~ s/{$key}/$context->{$key}/g;
  }

  if ($messageInfo->{'severity'} == WARNING) {
    Logging::warning("Pre-migration check: $briefMessage");
  } elsif ($messageInfo->{'severity'} == ERROR) {
    Logging::error("Pre-migration check: $briefMessage");
  }

  _addMessage($messageInfo->{'severity'}, $messageInfo->{'component'}, $briefMessage, $detailedMessage, $context, $owners);
}

# assert that some condition is true, and if so - add message
sub assert {
  my ($condition, $messageId, $context, $owners) = @_;

  if ($condition) {
    message($messageId, $context, $owners);
  }
}

sub printAllMessages {
  _printAllMessagesTo(*STDOUT);
}

sub writeBriefLog {
  my ($filename) = @_;
  
  if (open(my $fp, "> $filename")) {
    _printAllMessagesTo($fp);
    close($fp);
  } else {
    Logging::error("Failed to write brief pre-migration log to '$filename'");
  }
}

sub writeDetailedLog {
  my ($filename) = @_;

  my %fields = (
    'severityText' => 'Severity', 
    'component' => 'Component',
    'briefMessage' => 'Message',
    'detailedMessage' => 'Solution',
  );
  
  if (open(my $fp, "> $filename")) {
    foreach my $message (_getAllMessagesDefaultSorted()) {
      foreach my $key ('severityText', 'component', 'briefMessage', 'detailedMessage') {
        print $fp $fields{$key} . ': ' . $message->{$key} . "\n";
      }
      print $fp "--------------------------------------------------------------------------------\n";
    }
    close($fp);
  } else {
    Logging::error("Failed to write detailed pre-migration log to '$filename'");
  }
}

# @param $severity see the list of severity levels above
# @param $component see the list of components above
# @param $briefMessage brief description of the problem
# @param $detailedMessage detailed sescription of the problem, should contain a solution 
#   that could be used by customer to avoid or solve the problem
# @param $context reference to a hash with meta-information about context where the message was issued
#   for example, for domain it could be {'type' => 'domain', 'id' => 'example.com'}, for spamassassin option
#   it could be {'type' => 'spamassassion-option', 'name' => 'OPTION_NAME', 'value' => 'option_value', 'mailbox' => 'mail.example.tld'}
#   use this for filtering and sorting of the messages
# @param $owners reference to a list of object owners ('res0', 'web1', 'domain.com')
sub _addMessage {
  my ($severity, $component, $briefMessage, $detailedMessage, $context, $owners) = @_;

  my @owners = $owners ? @$owners{qw(anbieter kunde domain)} : ('', '', '');
  my @branch = map { defined $_ ? $_ : '' } @owners;
  push @MESSAGES, {
    'severity' => $severity, 
    'severityText' => SEVERITY_NAMES->{$severity},
    'component' => $component,
    'briefMessage' => $briefMessage,
    'detailedMessage' => $detailedMessage,
    'context' => $context,
    'owners' => \@branch,
  };
}

# TODO use createFulltree() , _walkTree() instead
sub _printAllMessagesTo {
  my ($fp) = @_;
    
  my @last_branch = ();
  foreach my $message (_getAllMessagesTreeSort()) {
    my ($indent, $path, $tail) = ('', '');
    my $is_common = "Common\n";
    my @branch = @{$message->{'owners'}};

    @last_branch = map { "_" } @branch unless @last_branch;
    for my $i (@branch) {
        my $leaf = '' eq $i ? $is_common : "-->$i\n";
        $is_common = '';
        $tail = shift @last_branch;
        if($i ne $tail) {
            # moved to adjacent branch, draw it
            $path .= ('' ne $i ? $indent : '').$leaf;
            $tail = $i;
            $is_common = '' ne $i && '_' ne $i ? "$indent   Common\n" : '';
        }
        $indent .= '    ';
        push @last_branch, $tail;
    }

    my $text = $message->{'severityText'} . "\n" . $message->{'briefMessage'} . "\n";
    $text =~ s/([^\n]{0,76})(?:\b\s*|\n)/$indent$1\n/gi;
    print $fp "$path$text";
  }
}

sub _getAllMessagesDefaultSorted {
  return sort { $b->{'severity'} <=> $a->{'severity'}} @MESSAGES;
}

# sort messages as tree; if severity<=> == 0, sort by reseller, then by cusromer, etc
sub _getAllMessagesTreeSort {
  return sort { 
      my $order = 0;
      if (@{$a->{'owners'}} && @{$b->{'owners'}} && @{$a->{'owners'}} == @{$b->{'owners'}} ) {
          for (my $i = 0; $i < @{$a->{'owners'}} ; $i++) {
              unless ($order) {
                  my ($left, $right) = (@{$a->{'owners'}}[$i], @{$b->{'owners'}}[$i]);  
                  if ($left gt $right ) {
                      $order = 1;
                  } elsif ($left lt $right ) {
                      $order = -1;
                  }
              }
          }
      }
      $order || $b->{'severity'} <=> $a->{'severity'}
  } @MESSAGES;

}

1;
